# incremental_obliterate.py - run obliterate in small incremental chunks
# Copyright robert@vaccaperna.co.uk

from __future__ import print_function

"""
Obliterate a Perforce path in an incremental fashion.
All the usual warnings about how dangerous this command is!!!

If you have largeish db.rev/integed tables (e.g. >10Gb) an obliterate can:
    - take large amounts of time and cause server thrashing depending on memory
    - end up taking a global lock on various tables which prevents other people working.

In this script we group things to gether and obliterate in reverse order of changelists.

Performance example - for 3k changelists, this took 30mins instead of hours and hours on a small server!

Also, it could be run during normal working hours (not peak load) as it executes lots of smaller
commands and doesn't lock the database.
"""

import P4
import time
import sys
import argparse

# Max no of files to obliterate together
MAX_FILES_PER_OBLITERATE = 5000

def message(*args):
    "Output to stdout with flush - useful for forked jobs"
    print(" ".join([str(arg) for arg in args]))
    sys.stdout.flush()
        
class HMS:
    "Output time (duration) in Hours:Minutes:Seconds"
    def __init__(self, secs):
        secs = int(secs)
        self.hours = secs / 3600
        secs = secs % 3600
        self.minutes = secs / 60
        secs = secs % 60
        self.seconds = secs
        
    def __repr__(self):
        return "%02d:%02d:%02d" % (
            self.hours, self.minutes, self.seconds)
            
class Status:
    "Utility class to output progress - performing rate calculations as appropriate"
    
    def __init__(self, total):
        self.total = total
        self.count = 0
        self.start_time = time.time()
        self.tick_start = time.time()

    def tick(self, count):
        "Called every action to output progress"
        self.count += count
        if self.count == 0:
            return
        elapsed = time.time() - self.tick_start
        self.tick_start = time.time()
        rate = (time.time() - self.start_time) / self.count
        time_togo = (self.total - self.count) * rate
        delta = HMS(time_togo)
        msg = "%s - so far/to go: %d / %d, %s / %s" % (
                str(HMS(elapsed)), 
                self.count, self.total - self.count, 
                str(HMS(time.time() - self.start_time)), str(delta))
        message(msg)
        sys.stdout.flush()

class Obliterater:
    
    def __init__(self, *args):
        parser = argparse.ArgumentParser(
            description="incremental_obliterate",
            epilog="Copyright (C) 2011-15 Robert Cowham"
        )

        parser.add_argument('-a', '--skip-archive', action='store_true', default='false', dest='skip_archive',
                    help="Skip archive file deletion")
        parser.add_argument('-b', '--only-branch', action='store_true', default='false', dest='only_branch', 
                    help="Only branched files (after the fact sparse branching)")
        parser.add_argument('-e', '--skip-have', action='store_true', default='false', dest='skip_have',
                    help="Skip deleting from db.have - equivalent of -h flag")
        parser.add_argument('-y', '--yes', action='store_true', default='false', 
                    help="Actually perform the obliterate instead of reporting")
        parser.add_argument('-p', '--port', help="P4PORT (otherwise takes it from the environment)")
        parser.add_argument('-u', '--user', help="P4USER (otherwise takes it from the environment)")
        parser.add_argument('path', help="Perforce depot path to obliterate (required)")

        self.options = parser.parse_args(list(args))
        self.p4 = P4.P4()
        if self.options.port:
            self.p4.port = self.options.port
        if self.options.user:
            self.p4.user = self.options.user
        message("options: ", self.options)
        message("Connecting to %s as user %s" % (self.p4.port, self.p4.user))
        self.p4.connect()

    def files_in_change(self, chgno):
        "Count how many files in the changelist"
        desc = self.p4.run_describe("-s", chgno)
        return len(desc[0]['depotFile'])
    
    def run_obl(self, path, grp):
        "Actually run a specific obliterate command on a group of changelists (reverse order)"
        if len(grp) == 0:
            return
        cmd = ["obliterate"]
        if self.options.skip_archive:
            cmd.append('-a')
        if self.options.only_branch:
            cmd.append('-b')
        if self.options.skip_have:
            cmd.append('-h')
        if self.options.yes:
            cmd.append('-y')
        cmd.append("%s@%s,@%s" % (path, grp[-1], grp[0]))
        result = self.p4.run(cmd)
        message("@%s,@%s %s" % (grp[-1], grp[0], result[-1]))
        for c in grp:
            self.p4.run_change("-f", "-d", c)
        
    def obl_path(self, path):
        "Obliterate all files in the path"
        changes = self.p4.run_changes(path)
        message("Obliterating %d changes for path %s" % (len(changes), path))
        status = Status(len(changes))
        counts = [] # array of tuples ('changeno', 'filecount')
        for c in changes:
            counts.append((c['change'], files_in_change(c['change'])))
    
        fcount = 0
        grp = []
        for chg in counts:
            if fcount + chg[1] > MAX_FILES_PER_OBLITERATE and len(grp) > 0:
                run_obl(path, grp)
                status.tick(len(grp))
                grp = []
                fcount = 0
            grp.append(chg[0])
            fcount += chg[1]
        if len(grp) > 0:    # don't forget final group if any
            run_obl(path, grp)
            status.tick(len(grp))

    def do_obliterate(self):
        "Perform the obliterate"
        for path in self.options.paths:
            self.obl_path(path)
 
def main():
    obl = Obliterater(*sys.argv[1:])
    obl.do_obliterate()
    
if __name__ == "__main__":
    main()
 